#!/usr/bin/env python3
#
# Copyright (c) 2022 Jan Behrens
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import random
import math

# helper function, which re-distributes counts such that they do not exceed
# nearest integers of the actual target values
def catpick_distribute(targets, lows, highs, high_prob):
    # calculate boundaries based on targets
    low_bounds = {}
    high_bounds = {}
    for cat, target in targets.items():
        low_bounds[cat] = math.floor(target)
        high_bounds[cat] = math.ceil(target)
    # categories which can still be adjusted
    open_cats = set(targets)
    while True:
        # errors (positive numbers) introduced in lows or highs
        low_error = 0
        high_error = 0
        # categories to be removed from open_cats after iteration
        cats_to_close = []
        # limit counts in categories which may still be adjusted
        # and remember errors caused
        for cat in open_cats:
            error = low_bounds[cat] - lows[cat]
            if error > 0:
                lows[cat] = low_bounds[cat]
                low_error += error
                other_error = error * ((1-high_prob) / high_prob)
                highs[cat] -= other_error
                high_error += other_error
                cats_to_close.append(cat)
                continue
            error = highs[cat] - high_bounds[cat]
            if error > 0:
                highs[cat] = high_bounds[cat]
                high_error += error
                other_error = error * (high_prob / (1-high_prob))
                lows[cat] += other_error
                low_error += other_error
                cats_to_close.append(cat)
        # mark categories which were force-limited as done
        for cat in cats_to_close:
            open_cats.remove(cat)
        # stop when all categories are done
        if not open_cats: break
        # also stop when there is no more error to be distributed
        if low_error == 0 and high_error == 0: break
        # distribute error
        low_total = 0
        high_total = 0
        for cat in open_cats:
            low_total += lows[cat]
            high_total += highs[cat]
        for cat in open_cats:
            if low_total: lows[cat] -= low_error * lows[cat] / low_total
            if high_total: highs[cat] += high_error * highs[cat] / high_total

# recursive helper function
def catpick_recursive(level, category_mapper, pool, count, round_up):

    # check illegal count
    if not count >= 0: raise ValueError("count not zero or positive")

    # perform rounding and calculate rounding probability
    count_floor = math.floor(count)
    count_ceil = math.ceil(count)
    count_frac = count % 1
    count_rounded = count_ceil if round_up else count_floor

    # avoid division by zero
    if count_rounded == 0: return []

    # category of each item in pool and number of items per category
    item_cats = []
    totals = {}
    for item in pool:
        cat = category_mapper(level, item)
        item_cats.append(cat)
        totals[cat] = totals.get(cat, 0) + 1

    # set of all categories
    cats = frozenset(totals)

    # if all categories are None, simply pick a random number of items
    if cats == frozenset((None,)):
        return random.sample(pool, count_rounded)

    # total count of pool
    total = len(pool)

    # counts to pick per category
    # (real target, and counts in case of rounding count down or up)
    targets, lows, highs = {}, {}, {}
    for cat in cats:
        target = count * totals[cat] / total
        targets[cat] = target
        lows[cat] = target * count_floor / count
        highs[cat] = target * count_ceil / count

    # limit counts to nearest integer
    catpick_distribute(targets, lows, highs, count_frac)

    # target counts for categories based on demanded rounding
    if round_up: targets = highs
    else: targets = lows

    # round target counts to integers
    rounded_counts = {}
    done_counter = 0
    remaining_targets = dict(targets)
    while done_counter < count_rounded:
        cat, target = remaining_targets.popitem()
        if not remaining_targets:
            rounded_counts[cat] = math.floor(target + 0.5)
            break
        round_up_error = 1 - (target % 1)
        round_down_error = target % 1
        remaining_total = 0
        for remaining_cat, remaining_target in remaining_targets.items():
            remaining_total += remaining_target
        lows, highs = {}, {}
        for remaining_cat, remaining_target in remaining_targets.items():
            lows[remaining_cat] = (
                remaining_target -
                round_up_error * remaining_target / remaining_total
            )
            highs[remaining_cat] = (
                remaining_target +
                round_down_error * remaining_target / remaining_total
            )
        catpick_distribute(remaining_targets, lows, highs, round_up_error)
        if round_up_error > random.random():
            remaining_targets = highs
            rounded_count = math.floor(target)
        else:
            remaining_targets = lows
            rounded_count = math.ceil(target)
        rounded_counts[cat] = rounded_count
        done_counter += rounded_count
    for cat in remaining_targets:
        rounded_counts[cat] = 0

    # all picked items
    result = []

    # for each category
    for cat in cats:

        # items in category
        subpool = []
        for item, item_cat in zip(pool, item_cats):
            if item_cat == cat:
                subpool.append(item)

        # pick items from category
        result += catpick_recursive(
            level+1,
            category_mapper,
            subpool,
            targets[cat],
            rounded_counts[cat] > targets[cat]
        )

    # return all picked items
    return result

# picks a certain count of items from a pool and ensures random order,
# where category_mapper maps a level and an item to a category
def catpick(category_mapper, pool, count):
    result = catpick_recursive(
        0,
        category_mapper,
        pool,
        count,
        count % 1 > random.random()
    )
    random.shuffle(result)
    return result

if __name__ == "__main__":

    def assert_eq(left, right):
        if not abs(left - right) < 1e-12:
            raise AssertionError("testing assertion failed")
    def assert_range(value, low, high):
        if not (value >= low and value <= high):
            raise AssertionError("testing assertion failed")

    def category_mapper(level, item):
        result = item[level:level+1]
        if result == "": return None
        return result

    print("Running test #1...")
    targets = {"A": 0.3, "B": 0.5, "C": 0.9}
    lows = {"A": 0.3/1.7, "B": 0.5/1.7, "C": 0.9/1.7}
    highs = {"A": 0.3*2/1.7, "B": 0.5*2/1.7, "C": 0.9*2/1.7}
    catpick_distribute(targets, lows, highs, 0.7)
    assert_eq(lows["A"] + lows["B"] + lows["C"], 1)
    assert_eq(highs["A"] + highs["B"] + highs["C"], 2)
    assert_eq(0.3 * lows["A"] + 0.7 * highs["A"], 0.3)
    assert_eq(0.3 * lows["B"] + 0.7 * highs["B"], 0.5)
    assert_eq(0.3 * lows["C"] + 0.7 * highs["C"], 0.9)
    assert_range(lows["A"], 0, 1)
    assert_range(lows["B"], 0, 1)
    assert_range(lows["C"], 0, 1)
    assert_range(highs["A"], 0, 1)
    assert_range(highs["B"], 0, 1)
    assert_range(highs["C"], 0, 1)
    print("Test passed.")

    print("Running test #2...")
    for i in range(30):
        targets = {
            "A": 2 * (1 - random.random()),
            "B": 2 * (1 - random.random()),
            "C": 2 * (1 - random.random()),
        }
        total = targets["A"] + targets["B"] + targets["C"]
        total_floor = math.floor(total)
        total_ceil = math.ceil(total)
        high_prob = total % 1
        low_prob = 1 - high_prob
        ups = 0
        lows, highs = {}, {}
        for cat in targets:
            lows[cat] = targets[cat] * total_floor / total
            highs[cat] = targets[cat] * total_ceil / total
        chosen = catpick_distribute(targets, lows, highs, high_prob)
        assert_eq(lows["A"] + lows["B"] + lows["C"], total_floor)
        assert_eq(highs["A"] + highs["B"] + highs["C"], total_ceil)
        assert_eq(low_prob*lows["A"] + high_prob*highs["A"], targets["A"])
        assert_eq(low_prob*lows["B"] + high_prob*highs["B"], targets["B"])
        assert_eq(low_prob*lows["C"] + high_prob*highs["C"], targets["C"])
        assert_range(
            lows["A"],
            math.floor(targets["A"]),
            math.ceil(targets["A"])
        )
        assert_range(
            lows["B"],
            math.floor(targets["B"]),
            math.ceil(targets["B"])
        )
        assert_range(
            lows["C"],
            math.floor(targets["C"]),
            math.ceil(targets["C"])
        )
    print("Test passed.")

    print("Running test #3...")
    pool = [
        "A1ax", "A1ay", "A1bx", "A1by", "A1cx", "A1cy",
        "A2ax", "A2ay", "A2bx", "A2by", "A2cx", "A2cy",
        "A3ax", "A3ay", "A3bx", "A3by", "A3cx", "A3cy",
        "B1ax", "B1ay", "B1bx", "B1by", "B1cx", "B1cy",
        "B2ax", "B2ay", "B2bx", "B2by", "B2cx", "B2cy",
        "C1ax", "C1ay", "C1bx", "C1by", "C1cx", "C1cy",
    ]
    lower = 10000 / 6 - ((5/36) * 10000)**0.5 * 4
    upper = 10000 / 6 + ((5/36) * 10000)**0.5 * 4
    counts = {}
    for r in range(10000):
        result = catpick(category_mapper, pool, 6)
        if len(result) != 6: raise RuntimeError("test failed")
        cnt_a = len(list(filter(lambda x: x[0] == "A", result)))
        cnt_b = len(list(filter(lambda x: x[0] == "B", result)))
        cnt_c = len(list(filter(lambda x: x[0] == "C", result)))
        cnt_a1 = len(list(filter(lambda x: x[0:2] == "A1", result)))
        cnt_a2 = len(list(filter(lambda x: x[0:2] == "A2", result)))
        cnt_a3 = len(list(filter(lambda x: x[0:2] == "A3", result)))
        if cnt_a != 3: raise RuntimeError("test failed")
        if cnt_b != 2: raise RuntimeError("test failed")
        if cnt_c != 1: raise RuntimeError("test failed")
        if cnt_a1 != 1: raise RuntimeError("test failed")
        if cnt_a2 != 1: raise RuntimeError("test failed")
        if cnt_a3 != 1: raise RuntimeError("test failed")
        for item in result:
            counts[item] = counts.get(item, 0) + 1
    for item, count in counts.items():
        if count < lower or count > upper:
            raise RuntimeError("test failed (statistics)")
    print("Test passed.")

    print("Running test #4...")
    pool = [
        "A1", "A2", "A3",
        "B1", "B2", "B3",
        "C1", "C2", "C3", "C4",
    ]
    counts = {}
    for r in range(10000):
        result = catpick(category_mapper, pool, 2)
        if len(result) != 2: raise RuntimeError("test failed")
        cnt_a = len(list(filter(lambda x: x[0] == "A", result)))
        cnt_b = len(list(filter(lambda x: x[0] == "B", result)))
        cnt_c = len(list(filter(lambda x: x[0] == "C", result)))
        if cnt_a > 1: raise RuntimeError("test failed")
        if cnt_b > 1: raise RuntimeError("test failed")
        if cnt_c > 1: raise RuntimeError("test failed")
    print("Test passed.")
